# -*- coding: utf-8 -*-
"""
/***************************************************************************
    begin                :    04/10/17
    git sha              :    :%H$
    copyright            :    (C) 2017 by Germán Carrillo (BSF-Swissphoto)
                              (C) 2016 by OPENGIS.ch
    email                :    gcarrillo@linuxmail.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
import re

from qgis.core import QgsApplication, QgsRelation, QgsWkbTypes
from qgis.PyQt.QtCore import QCoreApplication, QLocale, QObject, pyqtSignal

from ..dataobjects.fields import Field
from ..dataobjects.layers import Layer
from ..dataobjects.legend import LegendGroup
from ..dataobjects.relations import Relation
from ..db_factory.db_simple_factory import DbSimpleFactory
from ..utils.qt_utils import slugify
from .config import BASKET_FIELDNAMES, IGNORED_FIELDNAMES, READONLY_FIELDNAMES
from .domain_relations_generator import DomainRelationGenerator


class Generator(QObject):
    """Builds Model Baker objects from data extracted from databases."""

    stdout = pyqtSignal(str)
    new_message = pyqtSignal(int, str)

    def __init__(
        self,
        tool,
        uri,
        inheritance,
        schema=None,
        pg_estimated_metadata=False,
        parent=None,
        mgmt_uri=None,
        consider_basket_handling=False,
    ):
        """
        Creates a new Generator objects.
        :param uri: The uri that should be used in the resulting project. If authcfg is used, make sure the mgmt_uri is set as well.
        :param mgmt_uri: The uri that should be used to create schemas, tables and query meta information. Does not support authcfg.
        :consider_basket_handling: Makes the specific handling of basket tables depending if schema is created with createBasketCol.
        """
        QObject.__init__(self, parent)
        self.tool = tool
        self.uri = uri
        self.mgmt_uri = mgmt_uri
        self.inheritance = inheritance
        self.schema = schema or None
        self.pg_estimated_metadata = pg_estimated_metadata

        self.db_simple_factory = DbSimpleFactory()
        db_factory = self.db_simple_factory.create_factory(self.tool)
        self._db_connector = db_factory.get_db_connector(mgmt_uri or uri, schema)
        self._db_connector.stdout.connect(self.print_info)
        self._db_connector.new_message.connect(self.append_print_message)
        self.basket_handling = consider_basket_handling and self.get_basket_handling()

        self._additional_ignored_layers = (
            []
        )  # List of layers to ignore set by 3rd parties

        self.collected_print_messages = []

    def print_info(self, text):
        self.stdout.emit(text)

    def print_messages(self):
        for message in self.collected_print_messages:
            self.new_message.emit(message["level"], message["text"])
        self.collected_print_messages.clear()

    def append_print_message(self, level, text):
        message = {"level": level, "text": text}

        if message not in self.collected_print_messages:
            self.collected_print_messages.append(message)

    def layers(self, filter_layer_list=[]):
        ignore_basket_tables = not self.basket_handling
        tables_info = self.get_tables_info_without_ignored_tables(ignore_basket_tables)
        layers = list()

        db_factory = self.db_simple_factory.create_factory(self.tool)

        layer_uri = db_factory.get_layer_uri(self.uri)
        layer_uri.pg_estimated_metadata = self.pg_estimated_metadata

        # When a table has multiple geometry columns, it will be loaded multiple times (supported e.g. by PostGIS).
        table_appearance_count = {}

        for record in tables_info:
            if self.schema:
                if record["schemaname"] != self.schema:
                    continue
            if filter_layer_list and record["tablename"] not in filter_layer_list:
                continue
            table_appearance_count[record["tablename"]] = (
                table_appearance_count.get(record["tablename"], 0) + 1
            )

        for record in tables_info:
            # When in PostGIS mode, leaving schema blank should load tables from
            # all schemas, except the ignored ones
            if self.schema:
                if record["schemaname"] != self.schema:
                    continue

            if filter_layer_list and record["tablename"] not in filter_layer_list:
                continue

            is_domain = (
                record.get("kind_settings") == "ENUM"
                or record.get("kind_settings") == "CATALOGUE"
            )
            is_attribute = bool(record.get("attribute_name"))
            is_structure = record.get("kind_settings") == "STRUCTURE"
            is_nmrel = record.get("kind_settings") == "ASSOCIATION"
            # only when the basked_handling is active we will consider the table as basket table (according to it's name)
            is_basket_table = self.basket_handling and (
                record.get("tablename") == self._db_connector.basket_table_name
            )
            is_dataset_table = self.basket_handling and (
                record.get("tablename") == self._db_connector.dataset_table_name
            )

            alias = record["table_alias"] if "table_alias" in record else None
            if not alias:
                short_name = None
                if is_domain and is_attribute:
                    short_name = ""
                    if "ili_name" in record and record["ili_name"]:
                        short_name = (
                            record["ili_name"].split(".")[-2]
                            + "_"
                            + record["ili_name"].split(".")[-1]
                        )
                else:
                    if (
                        table_appearance_count[record["tablename"]] > 1
                        and "geometry_column" in record
                    ):
                        # multiple layers for this table - append geometry column to name
                        fields_info = self.get_fields_info(record["tablename"])
                        for field_info in fields_info:
                            if field_info["column_name"] == record["geometry_column"]:
                                if (
                                    "fully_qualified_name" in field_info
                                    and field_info["fully_qualified_name"]
                                ):
                                    short_name = (
                                        field_info["fully_qualified_name"].split(".")[
                                            -2
                                        ]
                                        + " ("
                                        + field_info["fully_qualified_name"].split(".")[
                                            -1
                                        ]
                                        + ")"
                                    )
                                else:
                                    short_name = record["tablename"]
                    elif "ili_name" in record and record["ili_name"]:
                        match = re.search(r"([^\(]*).*", record["ili_name"])
                        if match.group(0) == match.group(1):
                            short_name = match.group(1).split(".")[-1]
                        else:
                            # additional brackets in the the name - extended layer in geopackage
                            short_name = (
                                match.group(1).split(".")[-2]
                                + " ("
                                + match.group(1).split(".")[-1]
                                + ")"
                            )
                alias = short_name

            display_expression = ""
            if is_basket_table:
                display_expression = "coalesce(attribute(get_feature('{dataset_layer_name}', '{tid}', dataset), 'datasetname') || ' (' || topic || ') ', coalesce( attribute(get_feature('{dataset_layer_name}', '{tid}', dataset), 'datasetname'), {tilitid}))".format(
                    tid=self._db_connector.tid,
                    tilitid=self._db_connector.tilitid,
                    dataset_layer_name=self._db_connector.dataset_table_name,
                )
            elif "ili_name" in record and record["ili_name"]:
                meta_attrs = self.get_meta_attrs(record["ili_name"])
                for attr_record in meta_attrs:
                    if attr_record["attr_name"] == "dispExpression":
                        display_expression = attr_record["attr_value"]

            coord_decimals = (
                record["coord_decimals"] if "coord_decimals" in record else None
            )
            coordinate_precision = None
            if coord_decimals:
                coordinate_precision = 1 / (10**coord_decimals)

            layer = Layer(
                layer_uri.provider,
                layer_uri.get_data_source_uri(record),
                record.get("tablename"),
                record.get("srid"),
                record.get("extent"),
                record.get("geometry_column"),
                QgsWkbTypes.parseType(record["type"]) or QgsWkbTypes.Unknown,
                alias,
                is_domain,
                is_structure,
                is_nmrel,
                display_expression,
                coordinate_precision,
                is_basket_table,
                is_dataset_table,
                record.get("ili_name"),
            )

            # Configure fields for current table
            fields_info = self.get_fields_info(record["tablename"])
            min_max_info = self.get_min_max_info(record["tablename"])
            value_map_info = self.get_value_map_info(record["tablename"])
            re_iliname = re.compile(r".*\.(.*)$")

            for fielddef in fields_info:
                column_name = fielddef["column_name"]
                fully_qualified_name = (
                    fielddef["fully_qualified_name"]
                    if "fully_qualified_name" in fielddef
                    else None
                )
                m = (
                    re_iliname.match(fully_qualified_name)
                    if fully_qualified_name
                    else None
                )

                alias = None
                if "column_alias" in fielddef:
                    alias = fielddef["column_alias"]
                if m and not alias:
                    alias = m.group(1)

                field = Field(column_name)
                field.alias = alias

                # Should we hide the field?
                hide_attribute = False

                if "fully_qualified_name" in fielddef:
                    fully_qualified_name = fielddef["fully_qualified_name"]
                    if fully_qualified_name:
                        meta_attrs_column = self.get_meta_attrs(fully_qualified_name)

                        for attr_record in meta_attrs_column:
                            if attr_record["attr_name"] == "hidden":
                                if attr_record["attr_value"] == "True":
                                    hide_attribute = True
                                    break

                if column_name in IGNORED_FIELDNAMES:
                    hide_attribute = True

                if not self.basket_handling and column_name in BASKET_FIELDNAMES:
                    hide_attribute = True

                field.hidden = hide_attribute

                if column_name in READONLY_FIELDNAMES:
                    field.read_only = True

                if column_name in min_max_info:
                    field.widget = "Range"
                    field.widget_config["Min"] = min_max_info[column_name][0]
                    field.widget_config["Max"] = min_max_info[column_name][1]
                    if "numeric_scale" in fielddef:
                        field.widget_config["Step"] = pow(
                            10, -1 * fielddef["numeric_scale"]
                        )
                    # field.widget_config['Suffix'] = fielddef['unit'] if 'unit' in fielddef else ''
                    if "unit" in fielddef and fielddef["unit"] is not None:
                        field.alias = "{alias} [{unit}]".format(
                            alias=alias or column_name, unit=fielddef["unit"]
                        )

                if column_name in value_map_info:
                    field.widget = "ValueMap"
                    field.widget_config["map"] = [
                        {val: val} for val in value_map_info[column_name]
                    ]

                if "attr_mapping" in fielddef and fielddef["attr_mapping"] == "ARRAY":
                    field.widget = "List"

                if "texttype" in fielddef and fielddef["texttype"] == "MTEXT":
                    field.widget = "TextEdit"
                    field.widget_config["IsMultiline"] = True

                data_type = self._db_connector.map_data_types(fielddef["data_type"])
                if "time" in data_type or "date" in data_type:
                    field.widget = "DateTime"
                    field.widget_config["calendar_popup"] = True

                    dateFormat = QLocale(QgsApplication.instance().locale()).dateFormat(
                        QLocale.ShortFormat
                    )
                    timeFormat = QLocale(QgsApplication.instance().locale()).timeFormat(
                        QLocale.ShortFormat
                    )
                    dateTimeFormat = QLocale(
                        QgsApplication.instance().locale()
                    ).dateTimeFormat(QLocale.ShortFormat)

                    if data_type == self._db_connector.QGIS_TIME_TYPE:
                        field.widget_config["display_format"] = timeFormat
                    elif data_type == self._db_connector.QGIS_DATE_TIME_TYPE:
                        field.widget_config["display_format"] = dateTimeFormat
                    elif data_type == self._db_connector.QGIS_DATE_TYPE:
                        field.widget_config["display_format"] = dateFormat

                db_factory.customize_widget_editor(field, data_type)

                if "default_value_expression" in fielddef:
                    field.default_value_expression = fielddef[
                        "default_value_expression"
                    ]

                if self.basket_handling and column_name in BASKET_FIELDNAMES:
                    default_basket_topic = slugify(
                        f"default_basket{'_' if layer.model_topic_name else ''}{layer.model_topic_name}"
                    )
                    field.default_value_expression = f"@{default_basket_topic}"

                if column_name == self._db_connector.tilitid:
                    # when there is a t_ili_tid it should be filled up when there is no OID defined in the model
                    if "oid_domain" not in fielddef or fielddef["oid_domain"] is None:
                        field.default_value_expression = "substr(uuid(), 2, 36)"

                if "enum_domain" in fielddef and fielddef["enum_domain"]:
                    field.enum_domain = fielddef["enum_domain"]

                layer.fields.append(field)

            layers.append(layer)

        self.print_messages()

        return layers

    def relations(self, layers, filter_layer_list=[]):
        relations_info = self.get_relations_info(filter_layer_list)
        layer_map = dict()
        for layer in layers:
            if layer.name not in layer_map.keys():
                layer_map[layer.name] = list()
            layer_map[layer.name].append(layer)
        relations = list()

        classname_info = [
            record["iliname"] for record in self.get_iliname_dbname_mapping()
        ]

        for record in relations_info:
            if (
                record["referencing_table"] in layer_map.keys()
                and record["referenced_table"] in layer_map.keys()
            ):
                for referencing_layer in layer_map[record["referencing_table"]]:
                    for referenced_layer in layer_map[record["referenced_table"]]:
                        relation = Relation()
                        relation.referencing_layer = referencing_layer
                        relation.referenced_layer = referenced_layer
                        relation.referencing_field = record["referencing_column"]
                        relation.referenced_field = record["referenced_column"]
                        relation.name = record["constraint_name"]
                        relation.strength = (
                            QgsRelation.Composition
                            if "strength" in record
                            and record["strength"] == "COMPOSITE"
                            else QgsRelation.Association
                        )
                        relation.cardinality_max = record.get("cardinality_max", None)

                        # For domain-class relations, if we have an extended domain, get its child name
                        child_name = None
                        if referenced_layer.is_domain:
                            # Get child name (if domain is extended)
                            fields = [
                                field
                                for field in referencing_layer.fields
                                if field.name == record["referencing_column"]
                            ]
                            if fields:
                                field = fields[0]
                                if (
                                    field.enum_domain
                                    and field.enum_domain not in classname_info
                                ):
                                    child_name = field.enum_domain
                        relation.child_domain_name = child_name

                        relations.append(relation)

        if self._db_connector.ili_version() == 3:
            # Used for ili2db version 3 relation creation
            domain_relations_generator = DomainRelationGenerator(
                self._db_connector, self.inheritance
            )
            (
                domain_relations,
                bags_of_enum,
            ) = domain_relations_generator.get_domain_relations_info(layers)
            relations = relations + domain_relations
        else:
            # Create the bags_of_enum structure
            bags_of_info = self.get_bags_of_info()
            bags_of_enum = {}
            for record in bags_of_info:
                for layer in layers:
                    if record["current_layer_name"] == layer.name:
                        new_item_list = [
                            layer,
                            record["cardinality_min"]
                            + ".."
                            + record["cardinality_max"],
                            layer_map[record["target_layer_name"]][0],
                            self._db_connector.tid,
                            self._db_connector.dispName,
                        ]
                        unique_current_layer_name = "{}_{}".format(
                            record["current_layer_name"], layer.geometry_column
                        )
                        if unique_current_layer_name in bags_of_enum.keys():
                            bags_of_enum[unique_current_layer_name][
                                record["attribute"]
                            ] = new_item_list
                        else:
                            bags_of_enum[unique_current_layer_name] = {
                                record["attribute"]: new_item_list
                            }
        return (relations, bags_of_enum)

    def generate_node(self, layers, node_name, item_properties):
        if item_properties.get("group"):
            node = LegendGroup(
                QCoreApplication.translate("LegendGroup", node_name),
                static_sorting=True,
            )
        else:
            node = Layer(alias=node_name)  # create dummy
            layers.append(node)
        return node

    def full_node(self, layers, item, path_resolver=lambda path: path):
        current_node = None
        if item and isinstance(item, dict):
            current_node_name = next(iter(item))
            # when the node exists, but there is no content we proceed with an empty dict
            item_properties = item.get(current_node_name, None) or {}
            if item_properties.get("group"):
                # get group
                current_node = self.generate_node(
                    layers, current_node_name, item_properties
                )

                # append properties
                current_node.expanded = item_properties.get("expanded", True)
                current_node.checked = item_properties.get("checked", True)
                current_node.mutually_exclusive = item_properties.get(
                    "mutually-exclusive", False
                )
                current_node.mutually_exclusive_child = item_properties.get(
                    "mutually-exclusive-child", -1
                )
                current_node.definitionfile = path_resolver(
                    item_properties.get("definitionfile")
                )

                # append child-nodes
                if "child-nodes" in item_properties:
                    for child_item in item_properties["child-nodes"]:
                        node = self.full_node(layers, child_item, path_resolver)
                        if node:
                            current_node.append(node)
            else:
                # get layer according to the tablename or iliname
                if "tablename" in item_properties or "iliname" in item_properties:
                    for layer in layers:
                        tablename = item_properties.get("tablename")
                        iliname = item_properties.get("iliname")
                        if (
                            layer.name == tablename
                            or iliname
                            and layer.ili_name == iliname
                        ):
                            # if a geometrycolumn is provided, then only consider layer if it's the right one
                            geometry_column = item_properties.get("geometrycolumn")
                            if (
                                geometry_column
                                and layer.geometry_column != geometry_column
                            ):
                                continue
                            # otherwise consider the layer and change the alias
                            layer.alias = current_node_name
                            current_node = layer
                            break

                if not current_node:
                    # get the layer according to the alias
                    for layer in layers:
                        if layer.alias == current_node_name:
                            current_node = layer
                            break

                if not current_node:
                    # get the layer according to the name
                    for layer in layers:
                        if layer.name == current_node_name:
                            current_node = layer
                            break

                if not current_node:
                    current_node = self.generate_node(
                        layers, current_node_name, item_properties
                    )

                # append properties
                current_node.expanded = item_properties.get("expanded", True)
                current_node.checked = item_properties.get("checked", True)
                current_node.featurecount = item_properties.get("featurecount", False)
                if "uri" in item_properties:
                    current_node.uri = item_properties.get("uri")
                if "provider" in item_properties:
                    current_node.provider = item_properties.get("provider")
                current_node.definitionfile = path_resolver(
                    item_properties.get("definitionfile")
                )
                current_node.qmlstylefile = path_resolver(
                    item_properties.get("qmlstylefile")
                )
        return current_node

    def legend(
        self,
        layers,
        ignore_node_names=None,
        layertree_structure=None,
        path_resolver=lambda path: path,
    ):
        legend = LegendGroup(
            QCoreApplication.translate("LegendGroup", "root"),
            ignore_node_names=ignore_node_names,
            static_sorting=layertree_structure is not None,
        )
        if layertree_structure:
            for item in layertree_structure:
                node = self.full_node(layers, item, path_resolver)
                if node:
                    legend.append(node)
        else:
            tables = LegendGroup(QCoreApplication.translate("LegendGroup", "tables"))
            domains = LegendGroup(QCoreApplication.translate("LegendGroup", "domains"))
            domains.expanded = False
            system = LegendGroup(QCoreApplication.translate("LegendGroup", "system"))
            system.expanded = False

            point_layers = []
            line_layers = []
            polygon_layers = []

            for layer in layers:
                if layer.geometry_column:
                    geometry_type = QgsWkbTypes.geometryType(layer.wkb_type)
                    if geometry_type == QgsWkbTypes.PointGeometry:
                        point_layers.append(layer)
                    elif geometry_type == QgsWkbTypes.LineGeometry:
                        line_layers.append(layer)
                    elif geometry_type == QgsWkbTypes.PolygonGeometry:
                        polygon_layers.append(layer)
                else:
                    if layer.is_domain:
                        domains.append(layer)
                    elif layer.name in [
                        self._db_connector.basket_table_name,
                        self._db_connector.dataset_table_name,
                    ]:
                        system.append(layer)
                    else:
                        tables.append(layer)

            for l in polygon_layers:
                legend.append(l)
            for l in line_layers:
                legend.append(l)
            for l in point_layers:
                legend.append(l)

            if not tables.is_empty():
                legend.append(tables)
            if not domains.is_empty():
                legend.append(domains)
            if not system.is_empty():
                legend.append(system)

        return legend

    def db_or_schema_exists(self):
        return self._db_connector.db_or_schema_exists()

    def metadata_exists(self):
        return self._db_connector.metadata_exists()

    def set_additional_ignored_layers(self, layer_list):
        self._additional_ignored_layers = layer_list

    def get_ignored_layers(self, ignore_basket_tables=True):
        return (
            self._db_connector.get_ignored_layers(ignore_basket_tables)
            + self._additional_ignored_layers
        )

    def get_tables_info(self):
        return self._db_connector.get_tables_info()

    def get_meta_attrs_info(self):
        return self._db_connector.get_meta_attrs_info()

    def get_meta_attrs(self, ili_name):
        return self._db_connector.get_meta_attrs(ili_name)

    def get_fields_info(self, table_name):
        return self._db_connector.get_fields_info(table_name)

    def get_tables_info_without_ignored_tables(self, ignore_basket_tables=True):
        tables_info = self.get_tables_info()
        ignored_layers = self.get_ignored_layers(ignore_basket_tables)
        new_tables_info = []
        for record in tables_info:
            if self.schema:
                if record["schemaname"] != self.schema:
                    continue

            if ignored_layers and record["tablename"] in ignored_layers:
                continue

            new_tables_info.append(record)

        return new_tables_info

    def get_min_max_info(self, table_name):
        return self._db_connector.get_min_max_info(table_name)

    def get_value_map_info(self, table_name):
        return self._db_connector.get_value_map_info(table_name)

    def get_relations_info(self, filter_layer_list=[]):
        return self._db_connector.get_relations_info(filter_layer_list)

    def get_bags_of_info(self):
        return self._db_connector.get_bags_of_info()

    def get_iliname_dbname_mapping(self):
        return self._db_connector.get_iliname_dbname_mapping()

    def get_basket_handling(self):
        return self._db_connector.get_basket_handling()



# -*- coding: utf-8 -*-
"""
/***************************************************************************
    begin                :    05/03/18
    git sha              :    :%H$
    copyright            :    (C) 2018 by Germán Carrillo (BSF-Swissphoto)
    email                :    gcarrillo@linuxmail.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

from qgis.core import (
    Qgis,
    QgsLayerTreeLayer,
    QgsLayerTreeNode,
    QgsMapLayer,
    QgsWkbTypes,
)

layer_order = [
    "point",  # QgsWkbTypes.PointGeometry
    "line",  # QgsWkbTypes.LineGeometry
    "polygon",  # QgsWkbTypes.PolygonGeometry
    "raster",  # QgsMapLayer.RasterLayer
    "table",  # QgsWkbTypes.NullGeometry
    "unknown",
]  # Anything else like geometry collections or plugin layers


def get_first_index_for_layer_type(layer_type, group, ignore_node_names=None):
    """
    Finds the first index (from top to bottom) in the layer tree where a
    specific layer type is found. This function works only for the given group.
    """
    if ignore_node_names is None:
        ignore_node_names = []

    tree_nodes = group.children()

    for current, tree_node in enumerate(tree_nodes):
        if tree_node.name() in ignore_node_names:
            # Some groups (e.g., validation errors) might be useful on a
            # specific position (e.g., at the top), so, skip it to get indices
            continue
        if (
            tree_node.nodeType() == QgsLayerTreeNode.NodeGroup
            and layer_type != "unknown"
        ):
            return None
        elif (
            tree_node.nodeType() == QgsLayerTreeNode.NodeGroup
            and layer_type == "unknown"
        ):
            # We've reached the lowest index in the layer tree before a group
            return current

        # Make this check because children() sometimes returns nodes of type QgsLayerTreeNode instead.
        # This is a workaround for a weird behavior of QGIS.
        if isinstance(tree_node, QgsLayerTreeLayer):
            layer = tree_node.layer()
            if get_layer_type(layer) == layer_type:
                return current

    return None


def get_suggested_index_for_layer(layer, group, ignore_node_names=None):
    """
    Returns the index where a layer can be inserted, taking other layer types
    into account. For instance, if a line layer is given, this function will
    return the index of the first line layer in the layer tree (if above it
    there are no lower-ordered layers like tables, otherwise this returns that
    table index), and if there are no line layers in it, it will continue with
    the first index of polygons , rasters, tables, or groups. Always following
    the order given in the global layer_order variable.
    """
    indices = []
    index = None
    for layer_type in layer_order[
        layer_order.index(get_layer_type(layer)) :
    ]:  # Slice from current until last
        index = get_first_index_for_layer_type(layer_type, group, ignore_node_names)
        if index is not None:
            indices.append(index)

    if indices:
        index = min(indices)
    if index is None:
        return -1  # Send it to the last position in layer tree
    else:
        return index


def get_layer_type(layer):
    """
    To deal with all layer types, map their types to known values
    """
    if layer.type() == QgsMapLayer.VectorLayer:
        if layer.isSpatial():
            if layer.geometryType() == QgsWkbTypes.UnknownGeometry:
                return "unknown"
            else:
                return layer_order[layer.geometryType()]  # Point, Line or Polygon
        else:
            return "table"
    elif layer.type() == QgsMapLayer.RasterLayer:
        return "raster"
    else:
        return "unknown"


def get_group_non_recursive(group, group_name):
    groups = (
        group.findGroups(False)
        if Qgis.QGIS_VERSION_INT >= 31800
        else group.findGroups()
    )
    for group in groups:
        if group.name() == group_name:
            return group

    return None



from qgis.core import QgsRelation


class Relation(object):
    def __init__(self):
        self.referencing_layer = None
        self.referenced_layer = None
        self.referencing_field = None
        self.referenced_field = None
        self.name = None
        self.strength = QgsRelation.Association
        self.cardinality_max = None
        self.child_domain_name = None
        self.qgis_relation = None
        self._id = None

    def dump(self):
        definition = dict()
        definition["referencingLayer"] = self.referencing_layer
        definition["referencingField"] = self.referencing_field
        definition["referencedLayer"] = self.referenced_layer
        definition["referencedField"] = self.referenced_field
        definition["strength"] = self.strength
        definition["cardinality_max"] = self.cardinality_max
        definition["child_domain_name"] = self.child_domain_name

        return definition

    def load(self, definition):
        self.referencing_layer = definition["referencingLayer"]
        self.referencing_field = definition["referencingField"]
        self.referenced_layer = definition["referencedLayer"]
        self.referenced_field = definition["referencedField"]
        self.strength = definition["strength"]
        self.cardinality_max = definition["cardinality_max"]
        self.child_domain_name = definition["child_domain_name"]

    def create(self, qgis_project, relations):
        relation = QgsRelation()
        project_ids = qgis_project.relationManager().relations().keys()
        base_id = self.name

        suffix = 1
        self._id = base_id

        while self._id in project_ids:
            self._id = "{}_{}".format(base_id, suffix)
            suffix += 1

        while self._id in [rel.id() for rel in relations]:
            self._id = "{}_{}".format(base_id, suffix)
            suffix += 1

        relation.setId(self._id)
        relation.setName(self.name)
        relation.setReferencingLayer(self.referencing_layer.create().id())
        relation.setReferencedLayer(self.referenced_layer.create().id())
        relation.addFieldPair(self.referencing_field, self.referenced_field)
        relation.setStrength(self.strength)
        self.qgis_relation = relation
        return relation

    @property
    def id(self):
        return self._id



import configparser
import os

def conf_path() -> str:
    """
    Returns the path found for the pg_service.conf on the system as string.
    """
    pg_config_path = None
    if os.environ.get("PGSERVICEFILE"):
        pg_config_path = os.environ.get("PGSERVICEFILE")
    elif os.environ.get("PGSYSCONFDIR"):
        pg_config_path = os.path.join(os.environ.get("PGSYSCONFDIR"), "pg_service.conf")
    else:
        pg_config_path = "~/.pg_service.conf"
    return os.path.expanduser(pg_config_path)
 
def full_config( conf_file_path: str = None ) -> configparser.ConfigParser:
    """
    Returns full pgservice config as configparser.ConfigParser().
    """
    if conf_file_path is None:
        conf_file_path = conf_path()

    config = configparser.ConfigParser()
    if os.path.exists(conf_file_path):
        config.read(conf_file_path)
    return config

def service_config(service_name: str, conf_file_path: str = None) -> dict:
    """
    Returns the config from the given service name as a dict.
    """
    config = full_config(conf_file_path)
    if service_name in config:
        return dict(config[service_name])
    return {}

def write_service_setting(service_name: str, setting_key: str, setting_value: str, conf_file_path: str = None) -> bool:
    """
    Returns true if it could write the setting to the file
    """
    config = full_config(conf_file_path)
    if service_name in config:
        config[service_name][setting_key]=setting_value
        with open(conf_file_path or conf_path(), 'w') as configfile:
            config.write(configfile)
            return True
    return False

def service_names(conf_file_path: str = None) -> list:
    """
    Returns all service names in a list.
    """
    config = full_config(conf_file_path)
    return config.sections()



# -*- coding: utf-8 -*-
"""
/***************************************************************************
                              -------------------
        begin                : 2022-07-17
        git sha              : :%H$
        copyright            : (C) 2022 by Dave Signer
        email                : david at opengis ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
import os
import shutil

from .utils import slugify


class Target(object):
    """
    The target defines where to store the topping files (YAML, style, definition etc.)

    If there is no subdir it will look like:
    <maindir>
    ├── projecttopping
    │  └── <projectname>.yaml
    ├── layerstyle
    │  ├── <projectname>_<layername>.qml
    │  └── <projectname>_<layername>.qml
    └── layerdefinition
    │  └── <projectname>_<layername>.qlr

    With subdir:
    <maindir>
    ├── <subdir>
    │  ├── projecttopping
    │  │  └── <projectname>.yaml
    │  ├── layerstyle
    │  │  ├── <projectname>_<layername>.qml
    │  │  └── <projectname>_<layername>.qml
    │  └── layerdefinition
    │  │  └── <projectname>_<layername>.qlr
    """

    def __init__(
        self,
        projectname: str = "project",
        main_dir: str = None,
        sub_dir: str = None,
        path_resolver=None,
    ):
        self.projectname = projectname
        self.main_dir = main_dir
        self.sub_dir = sub_dir
        self.path_resolver = path_resolver

        if not path_resolver:
            self.path_resolver = self.default_path_resolver

        self.toppingfileinfo_list = []

    def filedir_path(self, file_dir):
        relative_path = os.path.join(self.sub_dir, file_dir)
        absolute_path = os.path.join(self.main_dir, relative_path)
        if not os.path.exists(absolute_path):
            os.makedirs(absolute_path)
        return absolute_path, relative_path

    def toppingfile_link(self, type: str, path: str):
        filename_slug = f"{slugify(self.projectname)}_{os.path.basename(path)}"
        absolute_filedir_path, relative_filedir_path = self.filedir_path(type)
        shutil.copy(
            path,
            os.path.join(absolute_filedir_path, filename_slug),
        )
        return self.path_resolver(self, filename_slug, type)

    @staticmethod
    def default_path_resolver(target, name, type):
        _, relative_filedir_path = target.filedir_path(type)

        toppingfile = {"path": os.path.join(relative_filedir_path, name), "type": type}
        target.toppingfileinfo_list.append(toppingfile)

        return os.path.join(relative_filedir_path, name)



# -*- coding: utf-8 -*-
"""
/***************************************************************************
    begin                :    08/04/19
    git sha              :    :%H$
    copyright            :    (C) 2019 by Yesid Polania
    email                :    yesidpol.3@gmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

from ..dbconnector.gpkg_connector import GPKGConnector
from .db_factory import DbFactory
from .gpkg_command_config_manager import GpkgCommandConfigManager
from .gpkg_layer_uri import GpkgLayerUri


class GpkgFactory(DbFactory):
    """Creates an entire set of objects so that QgisMmodelbakerodelBaker supports Geopackage database."""

    def get_db_connector(self, uri, schema):
        return GPKGConnector(uri, None)

    def get_db_command_config_manager(self, configuration):
        return GpkgCommandConfigManager(configuration)

    def get_layer_uri(self, uri):
        return GpkgLayerUri(uri)

    def pre_generate_project(self, configuration):
        return True, ""

    def post_generate_project_validations(self, configuration):
        return True, ""

    def get_specific_messages(self):
        messages = {"db_or_schema": "database", "layers_source": "GeoPackage"}

        return messages



# -*- coding: utf-8 -*-
"""
/***************************************************************************
    begin                :    30/04/19
    git sha              :    :%H$
    copyright            :    (C) 2019 by Yesid Polania
    email                :    yesidpol.3@gmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
from .layer_uri import LayerUri


class PgLayerUri(LayerUri):
    """Provides layer uri based on database uri (connection string) and specific information of the data source.

    This **layer uri** is used to create a Qgis layer.

    :ivar str uri: Database uri.
    """

    def __init__(self, uri):
        LayerUri.__init__(self, uri)
        self.pg_estimated_metadata = False
        self.provider = "postgres"

    def get_data_source_uri(self, record):
        if record["geometry_column"]:
            str_pg_estimated_metadata = (
                "true" if self.pg_estimated_metadata else "false"
            )
            data_source_uri = '{uri} key={primary_key} estimatedmetadata={estimated_metadata} srid={srid} type={type} table="{schema}"."{table}" ({geometry_column})'.format(
                uri=self.uri,
                primary_key=record["primary_key"],
                estimated_metadata=str_pg_estimated_metadata,
                srid=record["srid"],
                type=record["type"],
                schema=record["schemaname"],
                table=record["tablename"],
                geometry_column=record["geometry_column"],
            )
        else:
            data_source_uri = (
                '{uri} key={primary_key} table="{schema}"."{table}"'.format(
                    uri=self.uri,
                    primary_key=record["primary_key"],
                    schema=record["schemaname"],
                    table=record["tablename"],
                )
            )

        return data_source_uri



# -*- coding: utf-8 -*-
"""
/***************************************************************************
                              -------------------
        begin                : 2022-07-17
        git sha              : :%H$
        copyright            : (C) 2022 by Dave Signer
        email                : david at opengis ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
from enum import Enum
from typing import Union

from qgis.core import QgsLayerTreeGroup, QgsLayerTreeLayer


class ExportSettings(object):
    """
    The requested export settings of each node in the specific dicts:
    - qmlstyle_setting_nodes
    - definition_setting_nodes
    - source_setting_nodes

    The usual structure is using QgsLayerTreeNode as key and then export True/False
    {
        <QgsLayerTreeNode(Node1)>: { export: False }
        <QgsLayerTreeNode(Node2)>: { export: True }
    }

    But alternatively the layername can be used as key. In ProjectTopping it first looks up the node and if not available looking up the name.
    Using the node is much more consistent, since one can use layers with the same name, but for nodes you need the project already in advance.
    With name you can use prepared settings to pass (before the project exists) e.g. in automated workflows.
    {
        "Node1": { export: False }
        "Node2": { export: True }
    }

    For some settings we have additional info. Like in qmlstyle_nodes <QgsMapLayer.StyleCategories>. These are Flags, and can be constructed manually as well.
    qmlstyle_nodes =
    {
        <QgsLayerTreeNode(Node1)>: { export: False }
        <QgsLayerTreeNode(Node2)>: { export: True, categories: <QgsMapLayer.StyleCategories> }
    }
    """

    class ToppingType(Enum):
        QMLSTYLE = 1
        DEFINITION = 2
        SOURCE = 3

    def __init__(self):
        self.qmlstyle_setting_nodes = {}
        self.definition_setting_nodes = {}
        self.source_setting_nodes = {}

    def set_setting_values(
        self,
        type: ToppingType,
        node: Union[QgsLayerTreeLayer, QgsLayerTreeGroup] = None,
        name: str = None,
        export=True,
        categories=None,
    ) -> bool:
        """
        Appends the values (export, categories) to an existing setting
        """
        setting_nodes = self._setting_nodes(type)
        setting = self._get_setting(setting_nodes, node, name)
        setting["export"] = export
        if categories:
            setting["categories"] = categories
        return self._set_setting(setting_nodes, setting, node, name)

    def get_setting(
        self,
        type: ToppingType,
        node: Union[QgsLayerTreeLayer, QgsLayerTreeGroup] = None,
        name: str = None,
    ) -> dict():
        """
        Returns an existing or an empty setting dict
        """
        setting_nodes = self._setting_nodes(type)
        return self._get_setting(setting_nodes, node, name)

    def _setting_nodes(self, type: ToppingType):
        if type == ExportSettings.ToppingType.QMLSTYLE:
            return self.qmlstyle_setting_nodes
        if type == ExportSettings.ToppingType.DEFINITION:
            return self.definition_setting_nodes
        if type == ExportSettings.ToppingType.SOURCE:
            return self.source_setting_nodes

    def _get_setting(self, setting_nodes, node=None, name=None):
        setting = {}
        if node:
            setting = setting_nodes.get(node, {})
        if not setting:
            setting = setting_nodes.get(name, {})
        return setting

    def _set_setting(self, setting_nodes, setting, node=None, name=None) -> bool:
        if node:
            setting_nodes[node] = setting
            return True
        if name:
            setting_nodes[name] = setting
            return True
        return False



# -*- coding: utf-8 -*-
"""
/***************************************************************************
                              -------------------
        begin                : 2016-12-21
        git sha              : :%H$
        copyright            : (C) 2016 by OPENGIS.ch
        email                : info@opengis.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

from qgis.core import QgsLayerDefinition, QgsLayerTreeLayer, QgsProject

from ..utils.qgis_utils import get_group_non_recursive, get_suggested_index_for_layer


class LegendGroup(object):
    def __init__(
        self, name=None, expanded=True, ignore_node_names=None, static_sorting=False
    ):
        self.name = name
        self.items = list()
        self.expanded = expanded
        self.checked = True
        self.mutually_exclusive = False
        self.mutually_exclusive_child = -1
        self.definitionfile = None

        self.static_sorting = static_sorting

        # When adding layers in order, one could want to ignore nodes (e.g.,
        # groups that should be always on top)
        self.ignore_node_names = ignore_node_names

    def dump(self):
        definition = list()
        for item in self.items:
            definition.append(item.dump())
        return definition

    def append(self, item):
        self.items.append(item)

    def __getitem__(self, item):
        for i in self.items:
            try:
                if i.name == item:
                    return i
            except AttributeError:
                if i.table_name == item:
                    return i

        raise KeyError(item)

    def load(self, definition):
        self.items = definition

    def create(self, qgis_project: QgsProject, group=None):
        if not group:
            group = qgis_project.layerTreeRoot()

        existing_layer_source_uris = [
            found_layer.layer().dataProvider().dataSourceUri()
            for found_layer in qgis_project.layerTreeRoot().findLayers()
            if found_layer.layer().dataProvider()
        ]

        static_index = 0
        for item in self.items:
            if isinstance(item, LegendGroup):
                if item.definitionfile:
                    QgsLayerDefinition.loadLayerDefinition(
                        item.definitionfile, qgis_project, group
                    )
                else:
                    subgroup = get_group_non_recursive(group, item.name)
                    if subgroup is None:
                        subgroup = group.addGroup(item.name)
                    item.create(qgis_project, subgroup)
                    subgroup.setExpanded(item.expanded)
                    subgroup.setItemVisibilityChecked(item.checked)
                    subgroup.setIsMutuallyExclusive(
                        item.mutually_exclusive, item.mutually_exclusive_child
                    )
            else:
                layer = item.layer
                if (
                    not layer.dataProvider()
                    or layer.dataProvider().dataSourceUri()
                    not in existing_layer_source_uris
                ):
                    if self.static_sorting:
                        index = static_index
                    elif layer.isSpatial():
                        index = get_suggested_index_for_layer(
                            layer, group, self.ignore_node_names
                        )
                    else:
                        index = 0
                    layernode = QgsLayerTreeLayer(layer)
                    layernode.setExpanded(item.expanded)
                    layernode.setItemVisibilityChecked(item.checked)
                    layernode.setCustomProperty("showFeatureCount", item.featurecount)
                    group.insertChildNode(index, layernode)
            static_index += 1

    def is_empty(self):
        return not bool(self.items)



# -*- coding: utf-8 -*-
"""
/***************************************************************************
                              -------------------
        begin                : 14.9.17
        git sha              : :%H$
        copyright            : (C) 2017 by OPENGIS.ch
        email                : info@opengis.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
import glob
import os
import platform
import re
import subprocess
import tempfile
import zipfile

from qgis.PyQt.QtCore import QCoreApplication
from qgis.PyQt.QtGui import QColor

from ..utils.qt_utils import NetworkError, download_file
from .globals import DbIliMode
from .ili2dbtools import get_tool_url, get_tool_version


def get_ili2db_bin(tool, db_ili_version, stdout, stderr):
    if tool not in DbIliMode or tool == DbIliMode.ili:
        raise RuntimeError("Tool {} not found".format(tool))

    tool |= (
        DbIliMode.ili
    )  # Regardless of the incoming form (i.e., pg or ili2pg), we need its corresponding ili tool
    tool_name = tool.name  # in fact, we need the name of the ili tool

    ili_tool_version = get_tool_version(tool, db_ili_version)
    ili_tool_url = get_tool_url(tool, db_ili_version)

    dir_path = os.path.dirname(os.path.realpath(__file__))
    ili2db_dir = "{}-{}".format(tool_name, ili_tool_version)

    # the structure changed since 3.12.2
    if is_version_valid(
        ili_tool_version,
        "3.12.2",
        exact_required_version=False,
        module_tested=tool_name,
    ):
        ili2db_file = os.path.join(
            dir_path,
            "bin",
            ili2db_dir,
            "{tool}-{version}.jar".format(tool=tool_name, version=ili_tool_version),
        )
    else:
        ili2db_file = os.path.join(
            dir_path,
            "bin",
            ili2db_dir,
            "{tool}-{version}/{tool}.jar".format(
                tool=tool_name, version=ili_tool_version
            ),
        )

    if not os.path.isfile(ili2db_file):
        try:
            os.makedirs(os.path.join(dir_path, "bin", ili2db_dir), exist_ok=True)
        except FileExistsError:
            pass

        tmpfile = tempfile.NamedTemporaryFile(suffix=".zip", delete=False)

        stdout.emit(
            QCoreApplication.translate(
                "ili2dbutils",
                "Downloading {} version {}…".format(tool_name, ili_tool_version),
            )
        )

        try:
            download_file(
                ili_tool_url,
                tmpfile.name,
                on_progress=lambda received, total: stdout.emit("."),
            )
        except NetworkError as e:
            stderr.emit(
                QCoreApplication.translate(
                    "ili2dbutils",
                    'Could not download {tool_name}\n\n  Error: {error}\n\nFile "{file}" not found. Please download and extract <a href="{ili2db_url}">{tool_name}</a>'.format(
                        tool_name=tool_name,
                        ili2db_url=ili_tool_url,
                        error=e.msg,
                        file=ili2db_file,
                    ),
                )
            )
            return None

        try:
            with zipfile.ZipFile(tmpfile.name, "r") as z:
                z.extractall(os.path.join(dir_path, "bin", ili2db_dir))
        except zipfile.BadZipFile:
            # We will realize soon enough that the files were not extracted
            pass

        if not os.path.isfile(ili2db_file):
            stderr.emit(
                QCoreApplication.translate(
                    "ili2dbutils",
                    'File "{file}" not found. Please download and extract <a href="{ili2db_url}">{tool_name}</a>.'.format(
                        tool_name=tool_name, file=ili2db_file, ili2db_url=ili_tool_url
                    ),
                )
            )
            return None

    return ili2db_file


def get_all_modeldir_in_path(path, lambdafunction=None):
    all_subdirs = [path[0] for path in os.walk(path)]  # include path
    # Make sure path is included, it can be a special string like `%XTF_DIR`
    modeldirs = [path]
    for subdir in all_subdirs:
        if os.path.isdir(subdir) and len(glob.glob(subdir + "/*.ili")) > 0:
            if lambdafunction is not None:
                lambdafunction(subdir)
            modeldirs += [subdir]

    # Remove duplicates
    modeldirs = list(dict.fromkeys(modeldirs))
    return ";".join(modeldirs)


def color_log_text(text, txt_edit):
    textlines = text.splitlines()
    for textline in textlines:
        if textline.startswith("Warning:"):
            txt_edit.setTextColor(QColor("#FFBF00"))
            txt_edit.append(textline)
        elif "error" in textline.lower() or "failed" in textline.lower():
            txt_edit.setTextColor(QColor("#aa2222"))
            txt_edit.append(textline)
        else:
            txt_edit.setTextColor(QColor("#2a2a2a"))
            txt_edit.append(textline)


def get_java_path(base_configuration):
    """
    Delivers the path to a Java 8 installation or raises a JavaNotFoundError

    1. If the java path is set by the user, this one will be taken. No other paths are returned.
    2. Finds additional java paths in the enfironment variable JAVA_HOME (suffised with and without bin/java
    3. On Windows looks up the global system environment variable for PATH because this is mangled by
       the QGIS startup script.
    """

    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    # This function is like the linux which command
    def which(program):

        fpath, fname = os.path.split(program)
        if fpath:
            if is_exe(program):
                return program
        else:
            # for path in os.environ["PATH"].split(os.pathsep):
            for path in getenv_system("PATH").split(os.pathsep):
                path = path.strip('"')
                exe_file = os.path.join(path, program)
                if is_exe(exe_file):
                    return exe_file
        return None

    # This function gets a system variable
    # it was necessary to use this instead of os.environ["PATH"] because QGIS overwrites the path variable
    # the win32 libraries appear not to be part of the standard python install, but they are included in the
    # python version that comes with QGIS
    def getenv_system(varname, default=""):
        import winreg

        v = default
        try:
            rkey = winreg.OpenKey(
                winreg.HKEY_LOCAL_MACHINE,
                "SYSTEM\\CurrentControlSet\\Control\\Session Manager\\Environment",
            )
            try:
                v = str(winreg.QueryValueEx(rkey, varname)[0])
                v = winreg.ExpandEnvironmentStrings(v)
            except:
                pass
        finally:
            winreg.CloseKey(rkey)
        return v

    if base_configuration.java_path:
        # A java path is configured: respect it no mather what
        return base_configuration.java_path
    else:
        # By default try JAVA_HOME and PATH
        java_paths = []
        if "JAVA_HOME" in os.environ:
            paths = os.environ["JAVA_HOME"].split(";")
            for path in paths:
                # Include double check as java can be found on different paths
                # /usr/lib/jvm/java8oracle/bin/java
                java_paths += [
                    os.path.join(path.replace('"', "").replace("'", ""), "bin", "java")
                ]
                # C:\ProgramData\Oracle\Java\javapath\java
                java_paths += [
                    os.path.join(path.replace('"', "").replace("'", ""), "java")
                ]

        if platform.system() == "Windows":
            java_path = which("java.exe")
            if java_path:
                java_paths += [java_path]
        else:
            java_paths += ["java"]

        version_output = None
        java_version_re = re.compile(r'.*version "([0-9\.]+).*')
        for java_path in java_paths:
            try:
                startupinfo = None
                if platform.system() == "Windows":
                    startupinfo = subprocess.STARTUPINFO()
                    startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
                p = subprocess.Popen(
                    [java_path, "-version"],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    startupinfo=startupinfo,
                )
                output, err = p.communicate()
                version_output = err.decode("utf-8")
                java_version = java_version_re.match(version_output)
                if is_version_valid(
                    java_version.group(1),
                    "1.8.0",
                    exact_required_version=False,
                    module_tested="Java",
                ):
                    return java_path
            except FileNotFoundError:
                pass

        raise JavaNotFoundError(version_output)


def is_version_valid(
    current_version,
    min_required_version,
    exact_required_version=False,
    module_tested="",
):
    """
    Generic one, it helps us to validate whether a current version is greater or equal to a min_required_version or,
    if exact_required_version, if a current version is exactly the required one.

    Borrowed from 'Asistente LADM-COL'.

    :param current_version: String, in the form 2.9.5
    :param min_required_version: String, in the form 2.9.5
    :param exact_required_version: Boolean, if true, only the exact version is valid. If False, the min_required_version
                                   or any greater version will be accepted.
    :param module_tested: String, only for displaying a log with context
    :return: Whether the current version is valid or not
    """
    if current_version is None:
        return False

    current_version_splitted = current_version.split(".")
    if (
        len(current_version_splitted) < 4
    ):  # We could need 4 places for our custom plugin versions
        current_version_splitted = current_version_splitted + ["0", "0", "0", "0"]
        current_version_splitted = current_version_splitted[:4]

    min_required_version_splitted = min_required_version.split(".")
    if len(min_required_version_splitted) < 4:
        min_required_version_splitted = min_required_version_splitted + [
            "0",
            "0",
            "0",
            "0",
        ]
        min_required_version_splitted = min_required_version_splitted[:4]

    if exact_required_version:
        return min_required_version_splitted == current_version_splitted

    else:  # Min version and subsequent versions should work
        for i in range(len(current_version_splitted)):
            if int(current_version_splitted[i]) < int(min_required_version_splitted[i]):
                return False
            elif int(current_version_splitted[i]) > int(
                min_required_version_splitted[i]
            ):
                return True

    return True


class JavaNotFoundError(FileNotFoundError):
    """
    This error is raised when java could not be detected on the system.
    When the java_version variable is set, java (or any executable) was found
    but in the wrong version.
    If not, no executable was found at all.
    """

    def __init__(self, java_version=None):
        super().__init__()

        if java_version:
            self.java_version = java_version
        else:
            self.java_version = None

    @property
    def html_java_version(self):
        if self.java_version:
            return "<br/>".join(self.java_version.splitlines())
        else:
            return ""

    @property
    def error_string(self):
        if self.java_version:
            return QCoreApplication.translate(
                "ili2dbutils",
                'Wrong java version found. Qgis Model Baker requires at least java version 8. Please <a href="https://java.com/en/download/">install Java</a> and or <a href="#configure">configure a custom java path</a>.<br/><br/>Java Version:<br/>{}',
            ).format(self.html_java_version)
        else:
            return QCoreApplication.translate(
                "ili2dbutils",
                'Java 8 could not be found. Please <a href="https://java.com/en/download/">install Java</a> and or <a href="#configure">configure a custom java path</a>.',
            )



# -*- coding: utf-8 -*-
"""
/***************************************************************************
    begin                :    13/05/19
    git sha              :    :%H$
    copyright            :    (C) 2019 by Yesid Polania
    email                :    yesidpol.3@gmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
from qgis.PyQt.QtCore import QSettings

from ..libs import pgserviceparser
from .db_command_config_manager import DbCommandConfigManager


class PgCommandConfigManager(DbCommandConfigManager):
    """Manages a configuration object to return specific information of Postgres/Postgis.

    Provides database uri, arguments to ili2db and a way to save and load configurations parameters
    based on a object configuration.

    :ivar configuration object that will be managed
    """

    _settings_base_path = "ili2pg/"

    def __init__(self, configuration):
        DbCommandConfigManager.__init__(self, configuration)

    def get_uri(self, su=False, qgis=False):
        uri = []

        if su:
            uri += [
                "user={}".format(self.configuration.base_configuration.super_pg_user)
            ]
            if self.configuration.base_configuration.super_pg_password:
                uri += [
                    "password={}".format(
                        self.configuration.base_configuration.super_pg_password
                    )
                ]
            if self.configuration.sslmode:
                uri += ["sslmode='{}'".format(self.configuration.sslmode)]
            uri += ["host={}".format(self.configuration.dbhost)]
            if self.configuration.dbport:
                uri += ["port={}".format(self.configuration.dbport)]
            uri += ["dbname='{}'".format(self.configuration.database)]
            return " ".join(uri)

        if self.configuration.dbservice:
            uri += ["service='{}'".format(self.configuration.dbservice)]

        # only set the params when they are not available in the service
        if not pgserviceparser.service_config(self.configuration.dbservice).get(
            "sslmode", None
        ):
            if self.configuration.sslmode:
                uri += ["sslmode='{}'".format(self.configuration.sslmode)]

        if not pgserviceparser.service_config(self.configuration.dbservice).get(
            "host", None
        ):
            uri += ["host={}".format(self.configuration.dbhost)]

        if not pgserviceparser.service_config(self.configuration.dbservice).get(
            "port", None
        ):
            if self.configuration.dbport:
                uri += ["port={}".format(self.configuration.dbport)]

        if not pgserviceparser.service_config(self.configuration.dbservice).get(
            "dbname", None
        ):
            uri += ["dbname='{}'".format(self.configuration.database)]

        # only provide authcfg to the uri when it's needed for QGIS specific things
        if (
            qgis
            and self.configuration.dbauthid
            and not (
                pgserviceparser.service_config(self.configuration.dbservice).get(
                    "user", None
                )
                and pgserviceparser.service_config(self.configuration.dbservice).get(
                    "password", None
                )
            )
        ):
            uri += ["authcfg={}".format(self.configuration.dbauthid)]
        else:
            if not pgserviceparser.service_config(self.configuration.dbservice).get(
                "user", None
            ):
                uri += ["user={}".format(self.configuration.dbusr)]
            if not pgserviceparser.service_config(self.configuration.dbservice).get(
                "password", None
            ):
                if self.configuration.dbpwd:
                    uri += ["password={}".format(self.configuration.dbpwd)]

        return " ".join(uri)

    def get_db_args(self, hide_password=False, su=False):
        db_args = list()
        db_args += ["--dbhost", self.configuration.dbhost]
        if self.configuration.dbport:
            db_args += ["--dbport", self.configuration.dbport]
        if su:
            db_args += ["--dbusr", self.configuration.base_configuration.super_pg_user]
        else:
            db_args += ["--dbusr", self.configuration.dbusr]
        if (
            not su
            and self.configuration.dbpwd
            or su
            and self.configuration.base_configuration.super_pg_password
        ):
            if hide_password:
                db_args += ["--dbpwd", "******"]
            else:
                if su:
                    db_args += [
                        "--dbpwd",
                        self.configuration.base_configuration.super_pg_password,
                    ]
                else:
                    db_args += ["--dbpwd", self.configuration.dbpwd]
        db_args += ["--dbdatabase", self.configuration.database]
        db_args += [
            "--dbschema",
            self.configuration.dbschema or self.configuration.database,
        ]
        return db_args

    def get_schema_import_args(self):
        args = list()
        args += ["--setupPgExt"]
        return args

    def save_config_in_qsettings(self):
        settings = QSettings()
        # PostgreSQL specific options
        settings.setValue(self._settings_base_path + "host", self.configuration.dbhost)
        settings.setValue(self._settings_base_path + "port", self.configuration.dbport)
        settings.setValue(self._settings_base_path + "user", self.configuration.dbusr)
        settings.setValue(
            self._settings_base_path + "database", self.configuration.database
        )
        settings.setValue(
            self._settings_base_path + "schema", self.configuration.dbschema
        )
        settings.setValue(
            self._settings_base_path + "password", self.configuration.dbpwd
        )
        settings.setValue(
            self._settings_base_path + "usesuperlogin",
            self.configuration.db_use_super_login,
        )
        settings.setValue(
            self._settings_base_path + "authid", self.configuration.dbauthid
        )
        settings.setValue(
            self._settings_base_path + "service", self.configuration.dbservice
        )

    def load_config_from_qsettings(self):
        settings = QSettings()

        self.configuration.dbhost = settings.value(
            self._settings_base_path + "host", "localhost"
        )
        self.configuration.dbport = settings.value(self._settings_base_path + "port")
        self.configuration.dbusr = settings.value(self._settings_base_path + "user")
        self.configuration.database = settings.value(
            self._settings_base_path + "database"
        )
        self.configuration.dbschema = settings.value(
            self._settings_base_path + "schema"
        )
        self.configuration.dbpwd = settings.value(self._settings_base_path + "password")
        self.configuration.dbauthid = settings.value(
            self._settings_base_path + "authid"
        )
        self.configuration.dbservice = settings.value(
            self._settings_base_path + "service"
        )
        self.configuration.db_use_super_login = settings.value(
            self._settings_base_path + "usesuperlogin", defaultValue=False, type=bool
        )



# -*- coding: utf-8 -*-
"""
/***************************************************************************
    begin                :    30/04/19
    git sha              :    :%H$
    copyright            :    (C) 2019 by Yesid Polania
    email                :    yesidpol.3@gmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
from abc import ABC, abstractmethod


class LayerUri(ABC):
    """Provides layer uri based on database uri (connection string) and specific information of the data source. This is a abstract class.

    This **layer uri** is used to create a Qgis layer.

    :ivar str uri: Database uri.
    :ivar str provider: Database provider.
    """

    def __init__(self, uri):
        """
        :param str uri: Database uri. This is the same database uri of the db connectors.
        """
        self.uri = uri
        self.provider = None

    @abstractmethod
    def get_data_source_uri(self, record: dict):
        """Provides layer uri based on database uri and specific information of the data source.

        :param str record: Dictionary containing specific information of the data source.
        :return: Layer uri.
        :rtype: str
        """



IGNORED_SCHEMAS = ["pg_catalog", "information_schema"]

IGNORED_TABLES = [
    "spatial_ref_sys",
    "t_ili2db_import_object",
    "t_ili2db_attrname",
    "t_ili2db_settings",
    "t_ili2db_import",
    "t_ili2db_inheritance",
    "t_ili2db_model",
    "t_ili2db_classname",
    "t_ili2db_import_basket",
    "T_ILI2DB_TABLE_PROP",
    "T_ILI2DB_INHERITANCE",
    "T_ILI2DB_ATTRNAME",
    "T_ILI2DB_SETTINGS",
    "T_KEY_OBJECT",
    "T_ILI2DB_MODEL",
    "T_ILI2DB_IMPORT",
    "T_ILI2DB_TRAFO",
    "T_ILI2DB_COLUMN",
    "T_ILI2DB_COLUMN_PROP",
    "T_ILI2DB_CLASSNAME",
    "T_ILI2DB_IMPORT_OBJECT",
    "T_ILI2DB_IMPORT_BASKET",
    "T_ILI2DB_META_ATTRS",
    "ogr_empty_table",
    "gm_curve2dlistvalue",
    "gm_curve3dlistvalue",
    "gm_multicurve2d",
    "gm_multicurve3d",
    "gm_surface2dlistvalue",
    "gm_surface3dlistvalue",
    "gm_multisurface2d",
    "gm_multisurface3d",
    "gpkg_contents",
    "gpkg_data_column_constraints",
    "gpkg_data_columns",
    "gpkg_geometry_columns",
    "gpkg_extensions",
    "gpkg_spatial_ref_sys",
    "gpkg_ogr_contents",
    "gpkg_tile_matrix_set",
    "gpkg_tile_matrix",
    "gpkg_metadata",
    "gpkg_metadata_reference",
    "sqlite_sequence",
    "t_ili2db_basket",
    "T_ILI2DB_BASKET",
    "t_ili2db_dataset",
    "T_ILI2DB_DATASET",
]

BASKET_TABLES = [
    "t_ili2db_basket",
    "T_ILI2DB_BASKET",
    "t_ili2db_dataset",
    "T_ILI2DB_DATASET",
]

IGNORED_ILI_ELEMENTS = [
    "GeometryCHLV03_V1.MultiSurface",
    "GeometryCHLV03_V1.MultiLine",
    "GeometryCHLV03_V1.MultiDirectedLine",
    "GeometryCHLV95_V1.MultiSurface",
    "GeometryCHLV95_V1.MultiLine",
    "GeometryCHLV95_V1.MultiDirectedLine",
    "CatalogueObjects_V1.Catalogues.CatalogueReference",
    "CatalogueObjects_V1.Catalogues.MandatoryCatalogueReference",
    "LocalisationCH_V1.MultilingualMText",
    "LocalisationCH_V1.MultilingualText",
    "LocalisationCH_V1.LocalisedMText",
    "LocalisationCH_V1.LocalisedText",
]



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("VERSION", "r", encoding="utf-8") as fh:
    version = fh.read().strip()

setuptools.setup(
    name="modelbaker",
    version=version,
    author="Dave Signer",
    author_email="david@opengis.ch",
    description="The full model baker core",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/opengisch/QgisModelBakerLibrary",
    classifiers=[
        "Topic :: Database",
        "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    packages=setuptools.find_packages(exclude=["tests"]),
)



# -*- coding: utf-8 -*-
"""
/***************************************************************************
                              -------------------
        begin                : 2016-11-14
        git sha              : :%H$
        copyright            : (C) 2016 by OPENGIS.ch
        email                : info@opengis.ch
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
import logging

from qgis.core import (
    Qgis,
    QgsCoordinateReferenceSystem,
    QgsDataSourceUri,
    QgsExpressionContextUtils,
    QgsLayerDefinition,
    QgsRasterLayer,
    QgsRectangle,
    QgsVectorLayer,
    QgsWkbTypes,
)
from qgis.PyQt.QtCore import QCoreApplication, QSettings

from ..generator.config import IGNORED_FIELDNAMES
from .form import Form, FormFieldWidget, FormRelationWidget, FormTab


class Layer(object):
    def __init__(
        self,
        provider=None,
        uri=None,
        name=None,
        srid=None,
        extent=None,
        geometry_column=None,
        wkb_type=QgsWkbTypes.Unknown,
        alias=None,
        is_domain=False,
        is_structure=False,
        is_nmrel=False,
        display_expression=None,
        coordinate_precision=None,
        is_basket_table=False,
        is_dataset_table=False,
        ili_name=None,
        definitionfile=None,
        qmlstylefile=None,
    ):
        self.provider = provider
        self.uri = uri
        self.name = name
        if extent is not None:
            extent_coords = extent.split(";")
            extent = QgsRectangle(
                float(extent_coords[0]),
                float(extent_coords[1]),
                float(extent_coords[2]),
                float(extent_coords[3]),
            )
        self.extent = extent
        self.geometry_column = geometry_column
        self.wkb_type = wkb_type
        self.alias = alias
        self.__layer = None
        self.fields = list()
        self.is_domain = is_domain
        self.is_structure = is_structure

        self.is_nmrel = is_nmrel
        self.srid = srid
        """ If is_nmrel is set to true it is a junction table in a N:M relation.
        Or in ili2db terms, the table is marked as ASSOCIATION in t_ili2db_table_prop.settings. """

        self.display_expression = display_expression

        self.coordinate_precision = coordinate_precision

        self.is_basket_table = is_basket_table
        self.is_dataset_table = is_dataset_table

        self.ili_name = ili_name

        self.model_topic_name = ""
        if self.ili_name:
            if self.ili_name.count(".") > 1:
                self.model_topic_name = (
                    f"{self.ili_name.split('.')[0]}.{self.ili_name.split('.')[1]}"
                )

        self.definitionfile = definitionfile
        self.qmlstylefile = qmlstylefile

        self.__form = Form()

        # legend settings
        self.expanded = True
        self.checked = True
        self.featurecount = False

    def dump(self):
        definition = dict()
        definition["provider"] = self.provider
        definition["uri"] = self.uri
        definition["isdomain"] = self.is_domain
        definition["isstructure"] = self.is_structure
        definition["isnmrel"] = self.is_nmrel
        definition["isbaskettable"] = self.is_basket_table
        definition["isdatasettable"] = self.is_dataset_table
        definition["displayexpression"] = self.display_expression
        definition["coordinateprecision"] = self.coordinate_precision
        definition["modeltopicname"] = self.model_topic_name
        definition["ili_name"] = self.ili_name
        definition["definitionfile"] = self.definitionfile
        definition["qmlstylefile"] = self.qmlstylefile
        definition["form"] = self.__form.dump()
        return definition

    def load(self, definition):
        self.provider = definition["provider"]
        self.uri = definition["uri"]
        self.is_domain = definition["isdomain"]
        self.is_structure = definition["isstructure"]
        self.is_nmrel = definition["isnmrel"]
        self.is_basket_table = definition["isbaskettable"]
        self.is_dataset_table = definition["isdatasettable"]
        self.display_expression = definition["displayexpression"]
        self.coordinate_precision = definition["coordinateprecision"]
        self.model_topic_name = definition["modeltopicname"]
        self.ili_name = definition["ili_name"]
        self.definitionfile = definition["definitionfile"]
        self.qmlstylefile = definition["qmlstylefile"]
        self.__form.load(definition["form"])

    def create(self):
        if self.definitionfile:
            if self.__layer is None:
                layers = QgsLayerDefinition.loadLayerDefinitionLayers(
                    self.definitionfile
                )
                if layers:
                    self.__layer = layers[0]
            return self.__layer

        if self.__layer is None:
            layer_name = self.alias or self.name

            settings = QSettings()
            # Take the "CRS for new layers" config, overwrite it while loading layers and...
            old_proj_value = settings.value(
                "/Projections/defaultBehaviour", "prompt", type=str
            )
            settings.setValue("/Projections/defaultBehaviour", "useProject")
            self.__layer = self._create_layer(self.uri, layer_name, self.provider)
            settings.setValue("/Projections/defaultBehavior", old_proj_value)

            if (
                self.srid is not None
                and not self.__layer.crs().authid() == "EPSG:{}".format(self.srid)
            ):
                crs = QgsCoordinateReferenceSystem().fromEpsgId(self.srid)
                if not crs.isValid():
                    crs = QgsCoordinateReferenceSystem(self.srid)  # Fallback
                self.__layer.setCrs(crs)
            if self.is_domain or self.is_dataset_table:
                self.__layer.setReadOnly()
            if self.display_expression:
                self.__layer.setDisplayExpression(self.display_expression)
            if self.coordinate_precision and self.coordinate_precision < 1:
                self.__layer.geometryOptions().setGeometryPrecision(
                    self.coordinate_precision
                )
                self.__layer.geometryOptions().setRemoveDuplicateNodes(True)

            if self.model_topic_name:
                QgsExpressionContextUtils.setLayerVariable(
                    self.__layer, "interlis_topic", self.model_topic_name
                )

        for field in self.fields:
            field.create(self)

        return self.__layer

    def create_form(self, project):
        edit_form = self.__form.create(self, self.__layer, project)
        self.__layer.setEditFormConfig(edit_form)

    def load_style(self):
        if self.qmlstylefile:
            self.__layer.loadNamedStyle(self.qmlstylefile)

    def _create_layer(self, uri, layer_name, provider):
        if provider and provider.lower() == "wms":
            return QgsRasterLayer(uri, layer_name, provider)
        # return QgsVectorLayer even when it's an invalid layer with no provider
        return QgsVectorLayer(uri, layer_name, provider)

    def post_generate(self, project):
        """
        Will be called when the whole project has been generated and
        therefore all relations are available and the form
        can also be generated.
        """
        has_tabs = False
        for relation in project.relations:
            if relation.referenced_layer == self:
                has_tabs = True
                break

        if has_tabs:
            num_fields = len([f for f in self.fields if not f.hidden])
            if num_fields > 5:
                num_tabs = 2
            else:
                num_tabs = 1
            tab = FormTab(QCoreApplication.translate("FormTab", "General"), num_tabs)
            for field in self.fields:
                if not field.hidden:
                    widget = FormFieldWidget(field.alias, field.name)
                    tab.addChild(widget)

            self.__form.add_element(tab)

            relations_to_add = []
            for relation in project.relations:
                if relation.referenced_layer == self:

                    # 1:n relation will be added only if does not point to a pure link table
                    if (
                        not relation.referencing_layer.isPureLinkTable(project)
                        or Qgis.QGIS_VERSION_INT < 31600
                    ):
                        relations_to_add.append((relation, None))

                    else:
                        for nm_relation in project.relations:
                            if nm_relation == relation:
                                continue

                            if nm_relation.referenced_layer == self:
                                continue

                            # relations to the same table with different geometries should not be added
                            if (
                                self.srid
                                and nm_relation.referenced_layer.srid == self.srid
                            ):
                                continue

                            if nm_relation.referenced_layer.is_basket_table:
                                continue

                            if (
                                nm_relation.referencing_layer
                                == relation.referencing_layer
                            ):
                                relations_to_add.append((relation, nm_relation))

            for relation, nm_relation in relations_to_add:
                if nm_relation and Qgis.QGIS_VERSION_INT < 31600:
                    logger = logging.getLogger(__name__)
                    logger.warning(
                        'QGIS version older than 3.16. Relation editor widget for relation "{0}" will not be added.'.format(
                            nm_relation.name
                        )
                    )
                    continue

                if nm_relation:
                    tab = FormTab(nm_relation.referenced_layer.name)
                else:
                    tab = FormTab(relation.referencing_layer.name)

                widget = FormRelationWidget(relation, nm_relation)
                tab.addChild(widget)
                self.__form.add_element(tab)

        else:
            for field in self.fields:
                if not field.hidden:
                    widget = FormFieldWidget(field.alias, field.name)
                    self.__form.add_element(widget)

    def source(self):
        return QgsDataSourceUri(self.uri)

    @property
    def layer(self):
        return self.__layer

    @property
    def real_id(self):
        """
        The layer id. Only valid after creating the layer.
        """
        if self.__layer:
            return self.__layer.id()
        else:
            return None

    def isPureLinkTable(self, project):
        """
        Returns True if the layer is a pure link table in a n:m relation.
        With "pure" it is meant the layer has no more fields than foreign keys and its id.
        """

        remaining_fields = set()
        for field in self.fields:
            if field.name not in IGNORED_FIELDNAMES:
                remaining_fields.add(field.name)

        # Remove all fields that are referencing fields
        for relation in project.relations:
            if relation.referencing_layer != self:
                continue
            remaining_fields.discard(relation.referencing_field)

        return len(remaining_fields) == 0



# -*- coding: utf-8 -*-
"""
/***************************************************************************
                              -------------------
        begin                : 23/03/17
        git sha              : :%H$
        copyright            : (C) 2017 by Germán Carrillo (BSF-Swissphoto)
        email                : gcarrillo@linuxmail.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

from qgis.core import QgsNetworkAccessManager
from qgis.PyQt.QtNetwork import QNetworkProxy

from .ili2dbutils import get_all_modeldir_in_path


class BaseConfiguration(object):
    def __init__(self):
        self.super_pg_user = "postgres"
        self.super_pg_password = "postgres"

        self.custom_model_directories_enabled = False
        self.custom_model_directories = ""
        self.java_path = ""
        self.logfile_path = ""

        self.debugging_enabled = False

    def save(self, settings):
        settings.setValue("SuperUser", self.super_pg_user)
        settings.setValue("SuperPassword", self.super_pg_password)
        settings.setValue(
            "CustomModelDirectoriesEnabled", self.custom_model_directories_enabled
        )
        settings.setValue("CustomModelDirectories", self.custom_model_directories)
        settings.setValue("JavaPath", self.java_path)
        settings.setValue("LogfilePath", self.logfile_path)
        settings.setValue("DebuggingEnabled", self.debugging_enabled)

    def restore(self, settings):
        self.super_pg_user = settings.value("SuperUser", "postgres", str)
        self.super_pg_password = settings.value("SuperPassword", "postgres", str)
        self.custom_model_directories_enabled = settings.value(
            "CustomModelDirectoriesEnabled", False, bool
        )
        self.custom_model_directories = settings.value(
            "CustomModelDirectories", "", str
        )
        self.java_path = settings.value("JavaPath", "", str)
        self.debugging_enabled = settings.value("DebuggingEnabled", False, bool)
        self.logfile_path = settings.value("LogfilePath", "", str)

    def to_ili2db_args(self, with_modeldir=True):
        """
        Create an ili2db command line argument string from this configuration
        """
        args = list()

        if with_modeldir:
            if self.custom_model_directories_enabled and self.custom_model_directories:
                str_model_directories = [
                    get_all_modeldir_in_path(path)
                    for path in self.custom_model_directories.split(";")
                ]
                str_model_directories = ";".join(str_model_directories)
                args += ["--modeldir", str_model_directories]
        if self.debugging_enabled and self.logfile_path:
            args += ["--trace"]
            args += ["--log", self.logfile_path]
        return args

    @property
    def model_directories(self):
        dirs = list()
        if self.custom_model_directories_enabled and self.custom_model_directories:
            dirs = self.custom_model_directories.split(";")
        else:
            dirs = [
                "%ILI_FROM_DB",
                "%XTF_DIR",
                "http://models.interlis.ch/",
                "%JAR_DIR",
            ]
        return dirs

    @property
    def metaconfig_directories(self):
        dirs = list()
        if self.custom_model_directories_enabled and self.custom_model_directories:
            dirs = self.custom_model_directories.split(";")
        else:
            dirs = ["https://models.opengis.ch/"]
        return dirs


class Ili2DbCommandConfiguration(object):
    def __init__(self):
        self.base_configuration = BaseConfiguration()

        self.dbport = ""
        self.dbhost = ""
        self.dbpwd = ""
        self.dbusr = ""
        self.dbauthid = ""
        self.db_use_super_login = False
        self.database = ""
        self.dbschema = ""
        self.dbfile = ""
        self.dbservice = None
        self.sslmode = None
        self.tool = None
        self.ilifile = ""
        self.ilimodels = ""
        self.tomlfile = ""
        self.dbinstance = ""
        self.db_odbc_driver = ""
        self.disable_validation = False
        self.metaconfig = None
        self.metaconfig_id = None
        self.db_ili_version = None

    def append_args(self, args, values, consider_metaconfig=False):

        if consider_metaconfig and self.metaconfig and values:
            if "ch.ehi.ili2db" in self.metaconfig.sections():
                metaconfig_ili2db_params = self.metaconfig["ch.ehi.ili2db"]
                if values[0][2:] in metaconfig_ili2db_params.keys():
                    # since passing metaconfig is not supported yet. see https://github.com/claeis/ili2db/issues/392 we do not avoid adding the value to the args
                    print(
                        "Value {} from metaconfiguration file not considered. ".format(
                            values[0][2:]
                        )
                    )
        args += values

    def to_ili2db_args(self):

        # Valid ili file, don't pass --modeldir (it can cause ili2db errors)
        with_modeldir = not self.ilifile

        args = self.base_configuration.to_ili2db_args(with_modeldir=with_modeldir)

        proxy = QgsNetworkAccessManager.instance().fallbackProxy()
        if proxy.type() == QNetworkProxy.HttpProxy:
            self.append_args(args, ["--proxy", proxy.hostName()])
            self.append_args(args, ["--proxyPort", str(proxy.port())])

        if self.ilimodels:
            self.append_args(args, ["--models", self.ilimodels])

        if self.tomlfile:
            self.append_args(args, ["--iliMetaAttrs", self.tomlfile])

        return args


class ExportConfiguration(Ili2DbCommandConfiguration):
    def __init__(self):
        super().__init__()
        self.xtffile = ""
        self.with_exporttid = False
        self.iliexportmodels = ""
        self.dataset = ""
        self.baskets = list()

    def to_ili2db_args(self, extra_args=[], with_action=True):
        args = list()

        self.append_args(args, extra_args)

        if with_action:
            self.append_args(args, ["--export"])

        if self.disable_validation:
            self.append_args(args, ["--disableValidation"])

        if self.with_exporttid:
            self.append_args(args, ["--exportTid"])

        if self.iliexportmodels:
            self.append_args(args, ["--exportModels", self.iliexportmodels])

        if self.db_ili_version == 3:
            self.append_args(args, ["--export3"])

        if self.dataset:
            self.append_args(args, ["--dataset", self.dataset])

        if self.baskets:
            self.append_args(args, ["--baskets", ";".join(self.baskets)])

        self.append_args(args, Ili2DbCommandConfiguration.to_ili2db_args(self))

        self.append_args(args, [self.xtffile])

        return args


class SchemaImportConfiguration(Ili2DbCommandConfiguration):
    def __init__(self):
        super().__init__()
        self.inheritance = "smart1"
        self.create_basket_col = False
        self.create_import_tid = True
        self.srs_auth = "EPSG"  # Default SRS auth in ili2db
        self.srs_code = 2056  # Default SRS code in ili2db
        self.stroke_arcs = True
        self.pre_script = ""
        self.post_script = ""

    def to_ili2db_args(self, extra_args=[], with_action=True):
        """
        Create an ili2db argument array, with the password masked with ****** and optionally with the ``action``
        argument (--schemaimport) removed
        """
        args = list()

        if with_action:
            self.append_args(args, ["--schemaimport"])

        self.append_args(args, extra_args)

        self.append_args(args, ["--coalesceCatalogueRef"], True)
        self.append_args(args, ["--createEnumTabs"], True)

        if self.disable_validation:
            self.append_args(args, ["--sqlEnableNull"])

        else:
            self.append_args(args, ["--createNumChecks"])
            self.append_args(args, ["--createUnique"])
            self.append_args(args, ["--createFk"])

        self.append_args(args, ["--createFkIdx"], True)
        self.append_args(args, ["--coalesceMultiSurface"], True)
        self.append_args(args, ["--coalesceMultiLine"], True)
        self.append_args(args, ["--coalesceMultiPoint"], True)
        self.append_args(args, ["--coalesceArray"], True)
        self.append_args(args, ["--beautifyEnumDispName"], True)
        self.append_args(args, ["--createGeomIdx"], True)
        self.append_args(args, ["--createMetaInfo"], True)
        self.append_args(args, ["--expandMultilingual"], True)

        if self.db_ili_version is None or self.db_ili_version > 3:
            self.append_args(args, ["--createTypeConstraint"], True)
            self.append_args(args, ["--createEnumTabsWithId"], True)
            self.append_args(args, ["--createTidCol"], True)

        # version 3 backwards compatibility (not needed in newer versions)
        if self.create_import_tid:
            self.append_args(args, ["--importTid"])

        if self.inheritance == "smart1":
            self.append_args(args, ["--smart1Inheritance"])
        elif self.inheritance == "smart2":
            self.append_args(args, ["--smart2Inheritance"])
        else:
            self.append_args(args, ["--noSmartMapping"])

        if self.stroke_arcs:
            self.append_args(args, ["--strokeArcs"])

        if self.create_basket_col:
            self.append_args(args, ["--createBasketCol"])

        if self.srs_auth != "EPSG":
            self.append_args(args, ["--defaultSrsAuth", self.srs_auth])

        self.append_args(args, ["--defaultSrsCode", "{}".format(self.srs_code)])

        if self.pre_script:
            self.append_args(args, ["--preScript", self.pre_script])

        if self.post_script:
            self.append_args(args, ["--postScript", self.post_script])

        self.append_args(args, Ili2DbCommandConfiguration.to_ili2db_args(self))

        if self.ilifile:
            self.append_args(args, [self.ilifile])

        return args


class ImportDataConfiguration(SchemaImportConfiguration):
    def __init__(self):
        super().__init__()
        self.xtffile = ""
        self.delete_data = False
        self.with_importtid = False
        self.dataset = ""
        self.baskets = list()
        self.with_schemaimport = False

    def to_ili2db_args(self, extra_args=[], with_action=True):
        args = list()

        if with_action:
            self.append_args(args, ["--import"])

        if self.with_schemaimport:
            self.append_args(args, ["--doSchemaImport"])

        if self.disable_validation:
            self.append_args(args, ["--disableValidation"])

        if self.delete_data:
            self.append_args(args, ["--deleteData"])

        if self.with_importtid:
            self.append_args(args, ["--importTid"])

        if self.dataset:
            self.append_args(args, ["--dataset", self.dataset])

        if self.baskets:
            self.append_args(args, ["--baskets", ";".join(self.baskets)])

        self.append_args(
            args,
            SchemaImportConfiguration.to_ili2db_args(
                self, extra_args=extra_args, with_action=False
            ),
        )

        self.append_args(args, [self.xtffile])

        return args


class UpdateDataConfiguration(Ili2DbCommandConfiguration):
    def __init__(self):
        super().__init__()
        self.xtffile = ""
        self.dataset = ""
        self.with_importtid = False
        self.with_importbid = False

    def to_ili2db_args(self, extra_args=[], with_action=True):
        args = list()

        if with_action:
            self.append_args(args, ["--update"])

        self.append_args(args, extra_args)

        if self.disable_validation:
            self.append_args(args, ["--disableValidation"])

        if self.with_importtid:
            self.append_args(args, ["--importTid"])

        if self.with_importbid:
            self.append_args(args, ["--importBid"])

        self.append_args(args, ["--dataset", self.dataset])

        self.append_args(args, Ili2DbCommandConfiguration.to_ili2db_args(self))

        self.append_args(args, [self.xtffile])

        return args


class ValidateConfiguration(Ili2DbCommandConfiguration):
    def __init__(self):
        super().__init__()
        self.ilimodels = ""
        self.topics = ""
        self.dataset = ""
        self.baskets = list()
        self.with_exporttid = False
        self.xtflog = ""
        self.skip_geometry_errors = False
        self.valid_config = ""

    def to_ili2db_args(self, extra_args=[], with_action=True):
        args = list()

        if with_action:
            self.append_args(args, ["--validate"])

        self.append_args(args, extra_args)

        if self.ilimodels:
            self.append_args(args, ["--models", self.ilimodels])

        if self.topics:
            self.append_args(args, ["--topics", self.topics])

        if self.dataset:
            self.append_args(args, ["--dataset", self.dataset])

        if self.baskets:
            self.append_args(args, ["--baskets", ";".join(self.baskets)])

        if self.with_exporttid:
            self.append_args(args, ["--exportTid"])

        if self.xtflog:
            self.append_args(args, ["--xtflog", self.xtflog])

        if self.skip_geometry_errors:
            self.append_args(args, ["--skipGeometryErrors"])
            self.append_args(args, ["--disableAreaValidation"])

        if self.valid_config:
            self.append_args(args, ["--validConfig", self.valid_config])

        self.append_args(args, Ili2DbCommandConfiguration.to_ili2db_args(self))

        return args



# -*- coding: utf-8 -*-
"""
/***************************************************************************
    begin                :    08/04/19
    git sha              :    :%H$
    copyright            :    (C) 2019 by Yesid Polania
    email                :    yesidpol.3@gmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
from qgis.PyQt.QtCore import QCoreApplication

from ..dbconnector.db_connector import DBConnectorError
from ..dbconnector.pg_connector import PGConnector
from .db_factory import DbFactory
from .pg_command_config_manager import PgCommandConfigManager
from .pg_layer_uri import PgLayerUri


class PgFactory(DbFactory):
    """Creates an entire set of objects so that modelbaker supports Postgres/Postgis database."""

    def get_db_connector(self, uri, schema):
        return PGConnector(uri, schema)

    def get_db_command_config_manager(self, configuration):
        return PgCommandConfigManager(configuration)

    def get_layer_uri(self, uri):
        return PgLayerUri(uri)

    def pre_generate_project(self, configuration):
        result = not configuration.db_use_super_login
        message = ""

        if configuration.db_use_super_login:
            try:
                config_manager = self.get_db_command_config_manager(configuration)
                uri = config_manager.get_uri(True)

                _db_connector = self.get_db_connector(uri, configuration.dbschema)

                result = schema_exist = _db_connector.db_or_schema_exists()

                if not schema_exist:
                    _db_connector.create_db_or_schema(configuration.dbusr)
                    result = True

            except (DBConnectorError, FileNotFoundError) as e:
                message = QCoreApplication.translate(
                    "PgFactory",
                    "There was an error generating schema with superuser. Check superuser login parameters from settings > General.\n\nReason:\n\n{}",
                ).format(str(e))

        return result, message

    def post_generate_project_validations(self, configuration):
        result = False
        message = ""

        config_manager = self.get_db_command_config_manager(configuration)
        uri = config_manager.get_uri(configuration.db_use_super_login)

        connector = self.get_db_connector(uri, configuration.dbschema)

        if not connector._postgis_exists():
            message = QCoreApplication.translate(
                "PgFactory",
                "The current database does not have PostGIS installed! Please install it by running `CREATE EXTENSION postgis;` on the database before proceeding.",
            )
        else:
            result = True

        return result, message
